# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import jubatus

from .base import GenericSchema, BaseDataset, BaseService, GenericConfig
from .compat import *

class Schema(GenericSchema):
  """
  Schema for Weight service.
  """
  pass

class Dataset(BaseDataset):
  """
  Dataset for Weight service.
  """

  @classmethod
  def _predict(cls, row):
    return Schema.predict(row, False)

class Weight(BaseService):
  """
  Weight service.
  """

  @classmethod
  def name(cls):
    return 'weight'

  @classmethod
  def _client_class(cls):
    return jubatus.weight.client.Weight

  def update(self, dataset):
    """
    Updates the weight using the given dataset and returns extracted feature vectors.
    """

    cli = self._client()
    for (idx, d) in dataset:
      result = cli.update(d)
      yield (idx, result)

  def calc_weight(self, dataset):
    """
    Returns extracted feature vectors, without modifying the weight model.
    """

    cli = self._client()
    for (idx, d) in dataset:
      result = cli.calc_weight(d)
      yield (idx, result)

class Config(GenericConfig):
  """
  Configuration to run Weight service.
  """

  @classmethod
  def methods(cls):
    return [None]

  @classmethod
  def _default_method(cls):
    return None

  @classmethod
  def _default_parameter(cls, method):
    return None



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import sys
import itertools

PYTHON3 = sys.version_info >= (3, 0)

zip_longest = itertools.zip_longest if PYTHON3 else itertools.izip_longest
unicode_t = str if PYTHON3 else unicode
long_t = int if PYTHON3 else long

if not PYTHON3:
  # Python 2.x
  range = xrange
  zip = itertools.izip



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

from unittest import TestCase

try:
  import numpy as np
  from scipy.sparse import csr_matrix
except ImportError:
  pass

from jubakit.loader.sparse import SparseMatrixLoader

from .. import requireSklearn

@requireSklearn
class SparseMatrixLoaderTest(TestCase):
  def _create_matrix(self):
    """
    Create a sparse matrix:

    [[1, 0, 2],
     [0, 0, 3],
     [4, 5, 6]]
    """
    row = np.array([0, 0, 1, 2, 2, 2])
    col = np.array([0, 2, 2, 0, 1, 2])
    data = np.array([1, 2, 3, 4, 5, 6])
    return csr_matrix((data, (row, col)), shape=(3, 3))

  def test_simple(self):
    loader = SparseMatrixLoader(
      self._create_matrix(),
      ['k1','k2','k3']
    )
    idx = 0
    for row in loader:
      if idx == 0:
        self.assertEqual(set(['k1', 'k3']), set(row.keys()))
        self.assertEqual(1, row['k1'])
        self.assertEqual(2, row['k3'])
      elif idx == 1:
        self.assertEqual(set(['k3']), set(row.keys()))
        self.assertEqual(3, row['k3'])
      elif idx == 2:
        self.assertEqual(set(['k1', 'k2', 'k3']), set(row.keys()))
        self.assertEqual(4, row['k1'])
        self.assertEqual(5, row['k2'])
        self.assertEqual(6, row['k3'])
      else:
        self.fail('unexpected row: {0}'.format(idx))
      idx += 1



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import os
import threading

try:
  # Python 3
  import queue
except ImportError:
  # Python 2
  import Queue as queue

import tweepy.auth
from tweepy.streaming import StreamListener, Stream
from tweepy.auth import OAuthHandler
import jq

from ..base import BaseLoader
from ..compat import *

class TwitterStreamLoader(BaseLoader):
  """
  Loader to process Twitter Stream.
  Loads statuses only; other type of messages such as direct messages and
  warnings are just ignored.

  ``tweepy`` and ``jq`` package must be installed to use this loader.
  """

  # Keys frequently used are pre-defined.
  # You can optionally specify a list of custom keys (in `jq` selector syntax).
  STATUS_KEYS = [
    # Tweet
    '.id_str',                  # Unique ID
    '.text',                    # Content
    '.lang',                    # Language
    '.favorite_count',          # Num. of favs
    '.retweet_count',           # Num. of RTs
    '.timestamp_ms',            # Timestamp

    # User (the author of the tweet)
    '.user.id',                 # Unique ID
    '.user.name',               # Name
    '.user.screen_name',        # Account (without @ sign)
    '.user.description',        # Profile (bio)
    '.user.lang',               # Language
    '.user.statuses_count',     # Num. of tweets
    '.user.friends_count',      # Num. of followings
    '.user.followers_count',    # Num. of followers
    '.user.favourites_count',   # Num. of favs
    '.user.listed_count',       # Num. of lists author appears
  ]

  # Stream Modes:
  SAMPLE   = 'sample'
  FILTER   = 'filter'
  FIREHOSE = 'firehose'
  USER     = 'user'
  SITE     = 'site'

  def __init__(self, auth=None, mode=SAMPLE, keys=STATUS_KEYS, count=None, **kwargs):
    if auth is None:
      auth = TwitterOAuthHandler()

    self._listener = _TwitterStreamListener(self, keys)
    self._stream = tweepy.streaming.Stream(auth.get(), self._listener, secure=True)
    self._count = count
    self._queue = queue.Queue()

    start_stream = {
      self.SAMPLE:   self._stream.sample,
      self.FILTER:   self._stream.filter,
      self.FIREHOSE: self._stream.firehose,
      self.USER:     self._stream.userstream,
      self.SITE:     self._stream.sitestream,
    }.get(mode, None)

    if start_stream is None:
      raise RuntimeError('unknown stream mode: {0}'.format(mode))

    kwargs['async'] = False
    self._thread = threading.Thread(target=start_stream, kwargs=kwargs)
    self._thread.daemon = True

  def is_infinite(self):
    return self._count is None

  def _on_event(self, event):
    self._queue.put(event)

  def rows(self):
    self._thread.start()
    exception = None
    try:
      i = 0
      while True:
        self._listener.check_error()
        try:
          yield self._queue.get(True, 1)
        except queue.Empty:
          continue
        self._queue.task_done()
        i += 1
        if self._count is not None and self._count <= i:
          break
    finally:
      self._stream.disconnect()
      self._thread.join()

class _TwitterStreamListener(tweepy.streaming.StreamListener):
  def __init__(self, loader, keys):
    self._loader = loader
    self._keys = keys
    self._error = None
    super(_TwitterStreamListener, self).__init__()

  def check_error(self):
    if self._error:
      raise self._error

  def on_status(self, status):
    row = dict([(key, jq.jq(key).transform(status._json)) for key in self._keys])
    self._loader._on_event(row)

  def on_error(self, status_code):
    self._error = RuntimeError('Twitter Streaming API returned HTTP error {0}'.format(status_code))

  def on_exception(self, exception):
    self._error = exception

class TwitterOAuthHandler(object):
  """
  Handles authentication required to access Twitter Streaming API.
  """

  def __init__(self, **kwargs):
    """
    Authentication information must be specified as follows:

    >>> TwitterOAuth(
    ...   consumer_key='XXXXXXXXXXXXXXXXXXXX',
    ...   consumer_secret='XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX',
    ...   access_token='XXXXXXXX-XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX',
    ...   access_secret='XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX',
    ... )

    If some of keys are not specified, environmenet variables
    (TWITTER_CONSUMER_KEY etc.) will automatically be used.

    You can get your key by registering your app on: https://apps.twitter.com/
    """

    self._kwargs = kwargs

  def _v(self, key):
    if key in self._kwargs:
      return self._kwargs[key]
    envkey = 'TWITTER_{0}'.format(key.upper())
    if envkey in os.environ:
      return os.environ[envkey]
    raise RuntimeError('missing authentication information: {0} must be specified as a constructor argument or environment variable ({1})'.format(key, envkey))

  def get(self):
    auth = tweepy.auth.OAuthHandler(self._v('consumer_key'), self._v('consumer_secret'))
    auth.set_access_token(self._v('access_token'), self._v('access_secret'))
    return auth



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

from unittest import TestCase

from jubakit.loader.odbc import ODBCLoader

class ODBCLoaderTest(TestCase):
  def test_simple(self):
    loader = ODBCLoader()



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

from ..base import BaseLoader
from ..compat import *

class LineBasedStreamLoader(BaseLoader):
  """
  Loader to process line-oriented text stream.
  You can override `preprocess` method to separate the row into fields.
  """

  def __init__(self, f, close=False):
    self._f = f
    self._close = close
    self._lines = 0

  def rows(self):
    try:
      for line in self._f:
        yield {"number": self._lines, "line": line}
        self._lines += 1
    finally:
      if not self._f.closed and self._close:
        self._f.close()

class LineBasedFileLoader(LineBasedStreamLoader):
  """
  Loader to process line-oriented text file.
  """

  def __init__(self, filename, *args, **kwargs):
    f = open(filename, *args, **kwargs)
    super(LineBasedFileLoader, self).__init__(f, True)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import collections
import copy
import json
import random
import time
import logging
import subprocess
import tempfile

import jubatus
import msgpackrpc
import psutil

from .compat import *
from .logger import get_logger

_logger = get_logger()

class BaseLoader(object):
  """
  Loader loads rows from various data sources.
  """

  def is_infinite(self):
    """
    Returns True if the length of the data source is indeterminate (e.g., MQ.)
    """
    return False

  def preprocess(self, ent):
    """
    Preprocesses the given dict-like object into another dict-like object.
    The default implementation does not alter the object.  Users can override
    this method to perform custom process.  You can yield None to skip the
    record.
    """
    return ent

  def __iter__(self):
    """
    Loads each row from the data source.
    """
    for ent in self.rows():
      processed = self.preprocess(ent)
      if processed is not None:
        yield processed

  def rows(self):
    """
    Subclasses must override this method and yield each row of data source
    in flat dict-like object.  You can yield None to skip the record.
    """
    raise NotImplementedError()

class BaseSchema(object):
  """
  Schema defines data types for each key of the data.

  BaseSchema defines the fundamental 3 data types.

  - IGNORE: ignores the key (mainly intended for fallback)
  - AUTO: use the type of the key as its data type
  - INFER: guess the type of the key from its value; note that this is
           discouraged as it may result in unstable result.
  """

  # Data Types: single-character type names are reserved by jubakit.
  # External subclasses must use 2+ characters for type names.
  IGNORE = '_'
  AUTO = '.'
  INFER = '?'

  def __init__(self, mapping, fallback=None):
    """
    Defines a Schema.  Schema is an immutable object and cannot be modified.
    `mapping` is a dict-like object that maps row keys to the data type.
    Optionally you can assign an alias name for the key to handle different
    loaders with the same configuration.
    """
    self._fallback = fallback
    self._key2type = {}
    self._key2name = {}

    for (key, ent) in mapping.items():
      if isinstance(ent, (tuple, list, )):
        (key_type, key_name) = ent
      else:
        (key_type, key_name) = (ent, key)

      self._key2type[key] = key_type
      self._key2name[key] = key_name

  def transform(self, row):
    """
    Transform the row (dict-like) into data structures required by the
    corresponding Service.
    """
    raise NotImplementedError()

  @classmethod
  def predict(cls, row, typed):
    """
    Predicts a Schema from dict-like row object.
    """
    raise NotImplementedError()

  @staticmethod
  def _get_unique_mapping(mapping, fallback, key_type, name, optional=False):
    """
    Validates the schema key uniqueness.
    This is an utility method for subclasses.
    """
    if fallback == key_type:
      raise RuntimeError('{0} key cannot be specified as fallback in schema'.format(name))

    keys = [k for k in mapping.keys() if mapping[k] == key_type]
    if len(keys) == 0:
      if optional: return None
      raise RuntimeError('{0} key must be specified in schema'.format(name))
    elif 1 < len(keys):
      raise RuntimeError('{0} key must be an unique key in schema'.format(name))
    return keys[0]

  def __str__(self):
    return str({'keys': self._key2name, 'types': self._key2type, 'fallback_type': self._fallback})

class GenericSchema(BaseSchema):
  """
  GenericSchema is a base Schema class for all engines using Datum.

  GenericSchema defines 3 data types:

  - STRING: string features (string_values)
  - NUMBER: numeric features (num_values)
  - BINARY: binary features (binary_values)
  """

  STRING = 's'
  NUMBER = 'n'
  BINARY = 'b'

  def transform(self, row):
    """
    Transforms the row (represented in dict-like object) as Datum.
    Subclasses that define their own data types should override this method
    and handle them.
    """
    return self._transform_as_datum(row)

  def _add_to_datum(self, d, t, k, v):
    """
    Add value `v` whose type and name are `t` and `k` resp. to Datum `d`.
    """
    if v is None:
      return
    if t == self.STRING:
      d.add_string(k, unicode_t(v))
    elif t == self.NUMBER:
      if v != '':
        d.add_number(k, float(v))
    elif t == self.BINARY:
      d.add_binary(k, v)
    elif t == self.AUTO or t == self.INFER:
      (pred_type, pred_v) = self._predict_type(v, (t == self.AUTO))
      _logger.debug('key %s predicted as type %s', k, pred_type)
      self._add_to_datum(d, pred_type, k, pred_v)
    elif t == self.IGNORE:
      pass
    else:
      raise RuntimeError('invalid type {0} for key {1}'.format(t, k))

  def _transform_as_datum(self, row, d=None, skip_keys=[]):
    """
    Transforms the row as Datum.  If the original Datum `d` is specified,
    feature vectors will be added to it.
    """
    if d is None:
      d = jubatus.common.Datum()

    for (key, value) in row.items():
      if key in skip_keys:
        continue

      key_type = self._key2type.get(key, self._fallback)
      key_name = self._key2name.get(key, key)
      if key_type is None:
        raise RuntimeError('schema does not match: unknown key {0}'.format(key))
      self._add_to_datum(d, key_type, key_name, value)

    return d

  @classmethod
  def predict(cls, row, typed):
    """
    Predicts a schema from dict-like row object.
    """
    mapping = {}
    for (k, v) in row.items():
      (mapping[k], _) = cls._predict_type(v, typed)
      _logger.info('key %s predicted as type %s', k, mapping[k])
    return cls(mapping)

  @classmethod
  def _predict_type(cls, v, typed):
    """
    Predicts a data type for the given data.
    if `typed` is True, no type conversion will be tried against `v`.
    """
    if isinstance(v, (int, long_t, float)):
      return (cls.NUMBER, v)
    elif isinstance(v, unicode_t):
      if not typed:
        try: return (cls.NUMBER, float(v))
        except ValueError: pass
      return (cls.STRING, v)
    elif isinstance(v, bytes):
      if not typed:
        try: return (cls.NUMBER, float(v))
        except ValueError: pass
        try: return (cls.STRING, v.decode())
        except UnicodeDecodeError: pass
      return (cls.BINARY, v)
    else:
      raise ValueError('cannot detect data type of {0}'.format(v.__class__))

class BaseDataset(object):
  """
  Dataset is an abstract representation of set of data.
  """

  def __init__(self, loader, schema=None, static=None, _data=None):
    """
    Defines a new dataset.  Datasets are immutable and cannot be modified.

    Data will be loaded from the given `loader` using `schema`.

    When `static` is set to True (which is the default for non-infinite
    loaders), data will be loaded on memory immedeately; otherwise data
    will be loaded one-by-one from `loader`, which may be better when processing
    a large dataset.  For "infinite" loaders (like MQ and Twitter stream),
    `static` cannot be set to True.  Note that some features
    (e.g., index access) are not available for non-static datasets, which
    may be needed for some features like cross-validation etc.
    """
    self._loader = loader
    self._schema = schema
    self._index = -1
    self._buffer = None

    if static is None:
      # Non-infinite loaders are static by default.
      static = not loader.is_infinite()

    self._static = static

    # `_data` is internally used to create a shallow subset of the Dataset.
    if _data is None:
      self._data = []
    else:
      self._data = _data
      return  # the data is already loaded

    if static:
      if loader.is_infinite():
        # Infinite data sources (e.g., MQ) cannot be loaded statically on memory.
        raise RuntimeError('infinite loaders cannot be staticized')

      # Load all data entries.
      _logger.info('loading all records from loader %s', loader)
      for row in loader:
        # Predict schema.
        if self._schema is None:
          self._schema = self._predict(row)
        self._data.append(row)
      _logger.info('records loaded (%d entries)', len(self._data))

      # Don't hold a ref to the loader for static datasets.
      self._loader = None

  @classmethod
  def _predict(cls, row):
    """
    Predict the Schema for the given row using the corresponding Schema class.
    """
    # return GenericSchema.predict(row, False)
    raise NotImplementedError()

  def is_static(self):
    """
    Returns True for static datasets.
    """
    return self._static

  def get_schema(self):
    """
    Returns the Schema for this dataset.
    """
    return self._schema

  def shuffle(self, seed=None):
    """
    Returns a new immutable Dataset whose records are shuffled.
    """
    if not self._static:
      raise RuntimeError('non-static datasets cannot be shuffled')

    def _shuffle(data):
      return random.Random(seed).sample(data, len(data))
    return self.convert(_shuffle)

  def convert(self, func):
    """
    Applies the given callable (which is expected to perform batch
    pre-processing like `shuffle`) to the whole data entries and returns
    a new immutable Dataset.
    """
    if not self._static:
      raise RuntimeError('non-static datasets cannot be converted')
    new_data = func(self._data)
    if not isinstance(new_data, collections.Iterable):
      raise RuntimeError('convert function returned non-iterable: {0}'.format(new_data.__class__))

    return self.__class__(self._loader, self._schema, True, new_data)

  def get(self, idx):
    """
    Returns the raw entry loaded by Loader.
    """
    if idx == self._index:
      # For convenience, even non-static datasets can access the raw record for
      # the index that is currently being iterated.
      return self._buffer

    if not self._static:
      raise RuntimeError('non-static datasets cannot be random accessed by index')
    return self._data[idx]

  def __len__(self):
    """
    Returns the number of entries.
    """
    if not self._static:
      raise RuntimeError('length of non-static datasets cannot be retrieved')
    return len(self._data)

  def __getitem__(self, index):
    """
    Returns row(s) at the position `index`.
    `index` can be an iterable (like numpy array) or just an int.
    """
    if isinstance(index, int) and index == self._index:
      # For convenience, even non-static datasets can access the record for the
      # index that is currently being iterated.
      return self._schema.transform(self._buffer)

    if not self._static:
      raise RuntimeError('non-static datasets cannot be random accessed by index')

    if isinstance(index, slice):
      return self.__class__(self._loader, self._schema, True, self._data[index])
    elif isinstance(index, collections.Iterable):
      subdata = []
      for i in index:
        subdata.append(self._data[i])
      return self.__class__(self._loader, self._schema, True, subdata)
    else:
      return self._schema.transform(self._data[index])

  def __str__(self):
    return str(self._data)

  def __repr__(self):
    return repr(self._data)

  def __iter__(self):
    """
    Iteratively access each transformed rows.
    """
    try:
      source = self._data if self._static else self._loader
      self._index = 0
      for row in source:
        if row is None:
          # May contain None in self._data if Dataset.convert is used.
          continue
        self._buffer = row
        yield (self._index, self._schema.transform(row))
        self._index += 1
    finally:
      self._index = -1
      self._buffer = None
      self._loader = None

class BaseService(object):
  """
  Service provides an interface to machine learning features.
  """

  def __init__(self, host='127.0.0.1', port=9199, cluster='', timeout=0):
    """
    Creates a new service that connects to the exsiting server.
    """
    self._host = host
    self._port = port
    self._cluster = cluster
    self._timeout = timeout
    self._backend = None

  def __del__(self):
    """
    Destroys the backend process if exists.
    """
    backend = self._backend
    if backend is not None:
      backend.stop()

  @classmethod
  def name(cls):
    """
    Subclasses (Classifier, NearestNeighbor, ... etc.) must override this
    method and return its service name (classifier, nearest_neighbor, ... etc.)
    """
    #return 'classifier'
    raise NotImplementedError()

  @classmethod
  def _client_class(cls):
    """
    Subclasses must override this method and return the client class.
    """
    #return jubatus.classifier.client.Classifier
    raise NotImplementedError()

  @classmethod
  def run(cls, config, port=None):
    """
    Runs a new standalone server and returns the serivce instance to access
    the server.
    """
    backend = _ServiceBackend(cls.name(), config, port)
    _logger.info('service %s started on port %d', cls.name(), backend.port)

    # Returns the Service instance.
    service = cls('127.0.0.1', backend.port)
    service._backend = backend
    return service

  def _client(self):
    return self._client_class()(self._host, self._port, self._cluster, self._timeout)

  def stop(self):
    """
    Stops the backend process if exists.
    """
    if self._backend is not None:
      return self._backend.stop()

  def clear(self):
    """
    Clears the model.
    """
    if not self._client().clear():
      raise RuntimeError('failed to clear model')
    _logger.info('model cleared')

  def save(self, name, path=None):
    """
    Saves the model using `name`.  If `path` is specified, copy the saved
    model file to local `path`.
    """
    self._client().save(name)
    _logger.info('model saved: %s', name)
    # TODO copy source from `jubafetch` and make path option work.

  def load(self, name, path=None):
    """
    Loads the model using `name`.  If `path` is specified, copy the model
    file from local `path` to remote location.
    """
    if not self._client().load(name):
      raise RuntimeError('failed to load model: {0}'.format(name))
    _logger.info('model loaded: %s', name)
    # TODO copy source from `jubafetch` and make path option work.

  def get_status(self):
    """
    Returns the status of this server.  In distributed mode, returns statuses
    of all members.
    """
    return self._client().get_status()

class _ServiceBackend(object):
  """
  Service backend handles messy process-related things.
  """

  # Disable verbose tornado WARN logs on connect failure.
  logging.getLogger('tornado').setLevel(logging.ERROR)

  # Global Port-to-PID mapping.
  port2pid = {}

  def __init__(self, name, config, port=None):
    self.name = name
    self.config = config
    self.port = port
    self.log = None
    self._proc = None

    self._check_installed(name)

    if port is None:
      self.port = self._get_free_port()

    if port in self.port2pid:
      raise RuntimeError('port {0} currently in use by another service (PID {1})'.format(port, self.port2pid[port]))

    with tempfile.NamedTemporaryFile() as config_file:
      config_file.write(json.dumps(config).encode('utf-8'))
      config_file.flush()

      args = [
        'juba{0}'.format(name),
        '--listen_addr', '127.0.0.1',
        '--rpc-port', str(self.port),
        '--timeout', '0',
        '--configpath', config_file.name
      ]
      _logger.info('starting service: %s', args)
      self._proc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
      self._assign_port(self.port, self._proc.pid)

      # Wait until the RPC server start.
      started = self._wait_until_rpc_ready(self.port)
      if started:
        status = self.get_status()
        pid = int(status['pid'])
        if pid != self._proc.pid:
          self._proc.kill()
          raise RuntimeError('server cannot be started as port {0} conflicts with external Jubatus process (PID: {1})'.format(self.port, pid))

    if not started:
      _logger.error('failed to start service')
      log = self.stop()

  def __del__(self):
    proc = self._proc
    if proc is not None and proc.poll() is None:  # still running
      proc.kill()
      self._unassign_port(self.port, proc.pid)

  def stop(self):
    """
    Stops the server instance and return the server log.
    """
    proc = self._proc
    self._proc = None
    if proc is not None:
      if proc.poll() is None:  # still running
        _logger.debug('process is still running; will be terminated')
        proc.terminate()
      else:
        _logger.debug('process already terminated')

      _logger.debug('waiting for process to exit')
      (stdout, _) = proc.communicate()
      retval = proc.returncode
      _logger.debug('process exit with status %d', retval)
      self._unassign_port(self.port, proc.pid)
      if retval != 0:
        raise RuntimeError('server exit with status {0}; confirm that the config is valid: {1}'.format(retval, stdout))
      return stdout

  def get_status(self):
    cli = msgpackrpc.Client(msgpackrpc.Address('127.0.0.1', self.port), unpack_encoding='utf-8')
    try:
      return cli.call('get_status', '')['127.0.0.1_{0}'.format(self.port)]
    finally:
      cli.close()

  @classmethod
  def _get_free_port(cls, start=10000, end=30000):
    """
    Finds the free port available to listen.

    The default range of [10000,30000] is chosen to avoid the default
    ephemeral port range on most platforms.
    """
    used_ports = cls._get_ports_in_use()
    port = start
    while True:
      if port not in used_ports: return port
      port += 1
      if end < port:
        raise RuntimeError('no free port available in range [{0},{1}]'.format(start, end))

  @classmethod
  def _get_ports_in_use(cls):
    """
    Returns sorted list of ports currently used on localhost.
    """
    try:
      return sorted(set([x.laddr[1] for x in psutil.net_connections(kind='inet4')]))
    except psutil.AccessDenied:
      # On some platforms (such as OS X), root privilege is required to get used ports.
      # In that case we avoid port confliction to the best of our knowledge.
      _logger.info('ports in use cannot be obtained on this platform; ports will be assigned sequentially')
      return sorted(cls.port2pid.keys())

  @classmethod
  def _assign_port(cls, port, pid):
    m = cls.port2pid
    if port in m:
      raise RuntimeError('port {0} currently in use by PID {1}'.format(port, m[port]))
    m[port] = pid

  @classmethod
  def _unassign_port(cls, port, pid):
    m = cls.port2pid
    if m is None:
      pass   # maybe destruction is running.
    if port not in m:
      raise RuntimeError('port {0} is not in use'.format(port))
    if m[port] != pid:
      raise RuntimeError('port {0} is used by PID {1}, not PID {2}'.format(port, m[port], pcid))
    del m[port]

  @classmethod
  def _wait_until_rpc_ready(cls, port):
    sleep_time = 1000
    for i in range(10):
      time.sleep(sleep_time/1000000.0) # from usec to sec
      if cls._ping_rpc(port):
        _logger.debug('service RPC ready after %d tries', i)
        return True
      sleep_time *= 2
    return False

  @classmethod
  def _ping_rpc(cls, port):
    cli = msgpackrpc.Client(msgpackrpc.Address("127.0.0.1", port))
    try:
      cli.call('__ping__')
      raise AssertionError('dummy RPC succeeded')
    except msgpackrpc.error.RPCError as e:
      if e.args[0] == 1:  # "no such method"
        return True
    except:
      return False
    finally:
      cli.close()

    return False

  @classmethod
  def _check_installed(cls, name):
    procname = 'juba{0}'.format(name)

    _logger.debug('checking if service process %s is available', procname)
    try:
      proc = subprocess.Popen(
        [procname, '--version'],
        stdout=subprocess.PIPE, stderr=subprocess.STDOUT
      )
      (stdout, _) = proc.communicate()
      if proc.returncode == 0:
        return
      raise RuntimeError('{0} exit with status {1}; confirm that (DY)LD_LIBRARY_PATH is properly set: {2}'.format(procname, proc.returncode, stdout))
    except OSError as e:
      raise RuntimeError('{0} could not be started; confirm that PATH is properly set: {1}'.format(procname, e))

class BaseConfig(dict):
  """
  Config is a convenient class to build new config.
  """

  def __init__(self, *args, **kwargs):
    """
    Creates a new Config with default configuration.
    """
    super(BaseConfig, self).__init__(self)
    self._default(self)

  @classmethod
  def _default(cls, cfg):
    """
    Initializes the given config (dict-like) with the default configuration.
    """
    raise NotImplementedError()

  @classmethod
  def default(cls):
    """
    Returns a new default configuration.
    """
    cfg = {}
    cls._default(cfg)
    return cfg

class GenericConfig(BaseConfig):
  """
  GenericConfig is a base Config class for generic services that have
  `converter`, `method` and `parameter` in its config data.
  """

  _CONVERTER_TEMPLATE = {
    'string_filter_types' : {},
    'string_filter_rules' : [],
    'num_filter_types' : {},
    'num_filter_rules' : [],
    'string_types' : {},
    'string_rules' : [],
    'num_types': {},
    'num_rules': [],
    'binary_types': {},
    'binary_rules': [],
  }

  def __init__(self, method=None, parameter=None, converter=None):
    super(GenericConfig, self).__init__(self)

    if method is not None:
      self['method'] = method
      default_parameter = self._default_parameter(method)
      if default_parameter is None:
        if 'parameter' in self: del self['parameter']
      else:
        self['parameter'] = default_parameter

    if parameter is not None:
      if 'parameter' in self:
        self['parameter'].update(parameter)
      else:
        self['parameter'] = parameter

    if converter is not None:
      if 'converter' in self:
        self['converter'].update(converter)
      else:
        self['converter'] = converter

  @classmethod
  def _default(cls, cfg):
    cfg.clear()

    method = cls._default_method()
    parameter = cls._default_parameter(method)
    converter = cls._default_converter()

    if method    is not None: cfg['method'] = method
    if parameter is not None: cfg['parameter'] = parameter
    if converter is not None: cfg['converter'] = converter

  @classmethod
  def _default_method(cls):
    """
    Subclasses must override this method and return the preferred default
    method.
    """
    #return 'AROW'
    raise NotImplementedError()

  @classmethod
  def _default_parameter(cls, method):
    """
    Subclasses must override this method and return the preferred default
    parameter set for the specified method.  Return `None` if  the method
    does not require `parameter` block.
    """
    #return {'regularization_weight': 0.1}
    raise NotImplementedError()

  @classmethod
  def _default_converter(cls):
    """
    Returns a default converter contents.
    """
    cfg = copy.deepcopy(cls._CONVERTER_TEMPLATE)
    cfg['string_types'] = {
      'unigram': {'method': 'ngram', 'char_num': '1'},
      'bigram':  {'method': 'ngram', 'char_num': '2'},
      'trigram': {'method': 'ngram', 'char_num': '3'},
    }
    cfg['string_rules'] = [
      {'key': '*', 'type': 'unigram', 'sample_weight': 'tf', 'global_weight': 'idf'}
    ]
    cfg['num_rules'] = [
      {'key': '*', 'type': 'num'}
    ]

    return cfg

  @classmethod
  def methods(cls):
    """
    Subclasses must override this method and return methods available for
    this service.
    """
    #return ['perceptron', 'PA', 'AROW']
    raise NotImplementedError()

  def clear_converter(self):
    """
    Initialize the `converter` section of the config with an empty template.
    """
    self['converter'] = copy.deepcopy(self._CONVERTER_TEMPLATE)

  def add_mecab(self, name='mecab', arg='', ngram=1, base=False, include_features='*', exclude_features=''):
    """
    Add MeCab feature extraction to string_types.
    """
    if isinstance(include_features, list):
      include_features = '|'.join(include_features)

    if isinstance(exclude_features, list):
      exclude_features = '|'.join(exclude_features)

    self['converter']['string_types'][name] = {
      'method': 'dynamic',
      'path': 'libmecab_splitter.so',
      'function': 'create',
      'arg': arg,
      'ngram': str(ngram),
      'base': 'true' if base else 'false',
      'include_features': include_features,
      'exclude_features': exclude_features,
    }



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import csv
import io

from ..base import BaseLoader
from ..compat import *

class CSVLoader(BaseLoader):
  """
  Loader to process CSV files.
  """

  def __init__(self, filename, fieldnames=None, encoding='utf-8', *args, **kwargs):
    """
    Creates a new loader that processes CSV files.

    You can optionally give `fieldnames` option.
    If `fieldnames` is not specified (which is a default) or specifeid as True,
    the first line of the CSV is used for column names.  If `fieldnames` is
    specified as False, sequential column names are automatically generated
    like ['c0', 'c1', ...].  If `fieldnames` is a list, it is used as column
    names.
    """

    if fieldnames == True:
      # Automatically estimate field names later.
      fieldnames = None
    elif fieldnames == False:
      # Generate field names by peeking number of columns in the first row
      # of the CSV.
      with io.open(filename, encoding=encoding, newline='') as f:
        for ent in csv.reader(f):
          fieldnames = ['c{0}'.format(i) for i in range(len(ent))]
          break

    self._filename = filename
    self._fieldnames = fieldnames
    self._encoding = encoding
    self._args = args
    self._kwargs = kwargs

  def rows(self):
    def _encode_file(f, enc):
      for line in f: yield line.encode(enc)
    def _decode_row(r, enc):
      return dict([(k.decode(enc), r[k].decode(enc)) for k in r.keys()])

    with io.open(self._filename, encoding=self._encoding, newline='') as f:
      f = f if PYTHON3 else _encode_file(f, self._encoding)
      reader = csv.DictReader(f, fieldnames=self._fieldnames, *self._args, **self._kwargs)
      for row in reader:
        ent = row if PYTHON3 else _decode_row(row, self._encoding)
        yield ent



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

from ..base import BaseLoader
from ..compat import *

class ODBCLoader(BaseLoader):
  """
  Loader to process ODBC data sources.
  """

  # TODO not implemented yet.
  pass



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

from unittest import TestCase

from jubakit.loader.array import ArrayLoader, ZipArrayLoader

class ArrayLoaderTest(TestCase):
  def test_simple(self):
    loader = ArrayLoader(
      [['1','2'], ['3','4']],
      ['k1','k2']
    )
    for row in loader:
      self.assertEqual(set(['k1','k2']), set(row.keys()))
      if row['k1'] == '1':
        self.assertEqual('2', row['k2'])
      elif row['k1'] == '3':
        self.assertEqual('4', row['k2'])
      else:
        self.fail('unexpected row: {0}'.format(row))

  def test_auto_feature_name(self):
    loader = ArrayLoader([['1','2']])
    for row in loader:
      self.assertEqual(set(['v0', 'v1']), set(row.keys()))
      self.assertEqual('1', row['v0'])
      self.assertEqual('2', row['v1'])

  def test_none(self):
    loader = ArrayLoader([['1', None, '3']])
    for row in loader:
      self.assertEqual(set(['v0', 'v2']), set(row.keys()))
      self.assertEqual('1', row['v0'])
      self.assertEqual('3', row['v2'])

class ZipArrayLoaderTest(TestCase):
  def test_simple(self):
    loader = ZipArrayLoader(
      [['1','2','3'], ['x', 'y', 'z']],
      ['k1', 'k2']
    )
    self._check_loader(loader)

  def test_kwargs(self):
    loader = ZipArrayLoader(
      k1=['1','2','3'],
      k2=['x', 'y', 'z'],
    )
    self._check_loader(loader)

  def test_mixed(self):
    loader = ZipArrayLoader(
      [['1','2','3']],
      ['k1'],
      k2=['x', 'y', 'z']
    )
    self._check_loader(loader)

  def test_none(self):
    loader = ZipArrayLoader(
      k1=['1', None, '3'],
      k2=[None, '2', None],
    )
    for row in loader:
      if 'k1' in row:
        self.assertTrue(row['k1'] in ('1', '3'))
        self.assertTrue('k2' not in row)
      elif 'k2' in row:
        self.assertTrue(row['k2'] in ('2'))
        self.assertTrue('k1' not in row)
      else:
        self.fail('error')

  def test_error(self):
    self.assertRaises(RuntimeError, ZipArrayLoader, [['1','2','3'], ['x', 'y', 'z']], ['k1'])

  def _check_loader(self, loader):
    for row in loader:
      self.assertEqual(set(['k1', 'k2']), set(row.keys()))
      if row['k1'] == '1':
        self.assertEqual('x', row['k2'])
      elif row['k1'] == '2':
        self.assertEqual('y', row['k2'])
      elif row['k1'] == '3':
        self.assertEqual('z', row['k2'])
      else:
        self.fail('unexpected row')



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

from unittest import TestCase
from io import StringIO
from tempfile import NamedTemporaryFile as TempFile

from jubakit.loader.core import LineBasedStreamLoader, LineBasedFileLoader

class LineBasedStreamLoaderTest(TestCase):
  def test_simple(self):
    data = 'hello\nworld'

    f = StringIO(data)
    loader = LineBasedStreamLoader(f, True)
    lines = []
    for line in loader:
      lines.append(line)

    self.assertEqual([{'line': 'hello\n', 'number': 0}, {'line': 'world', 'number': 1}], lines)

  def test_close(self):
    data = 'hello\nworld'

    f = StringIO(data)
    loader = LineBasedStreamLoader(f, False)  # do not close
    for _ in loader: pass

    self.assertFalse(f.closed)

class LineBasedFileLoaderTest(TestCase):
  def test_simple(self):
    data = 'hello\nworld'
    lines = []

    with TempFile() as f:
      f.write(data.encode())
      f.flush()
      loader = LineBasedFileLoader(f.name)

      for line in loader:
        lines.append(line)

    self.assertEqual([{'line': 'hello\n', 'number': 0}, {'line': 'world', 'number': 1}], lines)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import jubatus

from .base import GenericSchema, BaseDataset, BaseService, GenericConfig
from .loader.array import ArrayLoader, ZipArrayLoader
from .loader.sparse import SparseMatrixLoader
from .loader.chain import ValueMapChainLoader, MergeChainLoader
from .compat import *

class Schema(GenericSchema):
  """
  Schema for Classifier service.
  """

  LABEL = 'l'

  def __init__(self, mapping, fallback=None):
    self._label_key = self._get_unique_mapping(mapping, fallback, self.LABEL, 'LABEL')
    super(Schema, self).__init__(mapping, fallback)

  def transform(self, row):
    """
    Classifier schema transforms the row into Datum and its associated label.
    """
    label = row.get(self._label_key, None)
    if label is not None:
      label = unicode_t(label)
    d = self._transform_as_datum(row, None, [self._label_key])
    return (label, d)

  @classmethod
  def predict(cls, row, typed):
    raise RuntimeError('Classifier schema cannot be auto predicted')

class Dataset(BaseDataset):
  """
  Dataset for Classifier service.
  """

  @classmethod
  def _predict(cls, row):
    raise RuntimeError('Classifier schema cannot be auto predicted')

  @classmethod
  def _from_loader(cls, data_loader, labels, label_names, static):
    # Label is feeded with '_label' key from Loader.
    label_loader = ZipArrayLoader(_label=labels)
    if label_names is not None:
      label_loader = ValueMapChainLoader(label_loader, '_label', label_names)
    loader = MergeChainLoader(data_loader, label_loader)
    schema = Schema({'_label': Schema.LABEL}, Schema.NUMBER)
    return Dataset(loader, schema, static)

  @classmethod
  def from_array(cls, data, labels, feature_names=None, label_names=None, static=True):
    """
    Converts two arrays (data and its associated labels) to Dataset.

    Parameters
    ----------
    data : array of shape [n_samples, n_features]
    labels : array of shape [n_samples]
    feature_names : array of shape [n_features], optional
    label_names : array of shape [n_labels], optional
    """
    data_loader = ArrayLoader(data, feature_names)
    return cls._from_loader(data_loader, labels, label_names, static)

  @classmethod
  def from_matrix(cls, data, labels, feature_names=None, label_names=None, static=True):
    """
    Converts a sparse matrix data and its associated label array to Dataset.

    Parameters
    ----------

    data : scipy 2-D sparse matrix of shape [n_samples, n_features]
    labels : array of shape [n_samples]
    feature_names : array of shape [n_features], optional
    label_names : array of shape [n_labels], optional
    """
    data_loader = SparseMatrixLoader(data, feature_names)
    return cls._from_loader(data_loader, labels, label_names, static)

  def get_labels(self):
    """
    Returns labels of each record in the dataset.
    """

    if not self._static:
      raise RuntimeError('non-static datasets cannot fetch list of labels')

    for (idx, (label, d)) in self:
      yield label

class Classifier(BaseService):
  """
  Classifier service.
  """

  @classmethod
  def name(cls):
    return 'classifier'

  @classmethod
  def _client_class(cls):
    return jubatus.classifier.client.Classifier

  def train(self, dataset):
    """
    Trains the classifier using the given dataset.
    """

    cli = self._client()
    for (idx, (label, d)) in dataset:
      assert label is not None
      result = cli.train([jubatus.classifier.types.LabeledDatum(unicode_t(label), d)])
      assert result == 1
      yield (idx, label)

  def classify(self, dataset):
    """
    Classify the given dataset using this classifier.
    """

    cli = self._client()
    for (idx, (label, d)) in dataset:
      # Do classification for the record.
      result = cli.classify([d])
      assert len(result) == 1

      # Create the list of (label, score) desc sorted by score.
      label_score_sorted = [(ent.label, ent.score) for ent in sorted(result[0], key=lambda x: x.score, reverse=True)]

      # Note: label may become None.
      yield (idx, label, label_score_sorted)

  @classmethod
  def train_and_classify(cls, config, train_dataset, test_dataset, metric):
    """
    This is an utility method to perform bulk train-test.
    Run a classifier using the given config, train the classifier, classify
    using the classifier, then return the calculated metrics.
    """
    classifier = cls.run(config)

    for _ in classifier.train(train_dataset):
      pass

    y_true = []
    y_pred = []
    for (idx, label, result) in classifier.classify(test_dataset):
      if 0 < len(result):
        y_true.append(label)
        y_pred.append(result[0][0])

    classifier.stop()

    return metric(y_true, y_pred)

class Config(GenericConfig):
  """
  Configuration to run Classifier service.
  """

  @classmethod
  def methods(cls):
    return ['perceptron', 'PA', 'PA1', 'PA2', 'CW', 'AROW', 'NHERD', 'NN', 'cosine', 'euclidean']

  @classmethod
  def _default_method(cls):
    return 'AROW'

  @classmethod
  def _default_parameter(cls, method):
    if method in ('perceptron',
                  'PA', 'passive_aggressive'):
      return None
    elif method in ('PA1', 'passive_aggressive_1',
                    'PA2', 'passive_aggressive_2',
                    'CW', 'confidence_weighted',
                    'AROW',
                    'NHERD', 'normal_herd'):
      return {'regularization_weight': 1.0}
    elif method in ('cosine', 'euclidean'):
      return {
        'nearest_neighbor_num': 128,
        'local_sensitivity': 1.0,
      }
    elif method in ('NN', 'nearest_neighbor'):
      return {
        'method': 'euclid_lsh',
        'parameter': {
          'threads': -1,  # use number of logical CPU cores
          'hash_num': 64,
        },
        'nearest_neighbor_num': 128,
        'local_sensitivity': 1.0
      }
    else:
      raise RuntimeError('unknown method: {0}'.format(method))



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

from ..base import BaseLoader
from ..compat import *

class ArrayLoader(BaseLoader):
  """
  ArrayLoader is a loader to read from 2-d array.
  Expected to load row-oriented data.

  For example:

  >>> ArrayLoader([[1,2,3], [4,5,6]], ['k1','k2','k3'])

  ... will load two entries:

    - {'k1': 1, 'k2': 2, 'k3': 3}
    - {'k1': 4, 'k2': 5, 'k3': 6}
  """

  def __init__(self, array, feature_names=None):
    if feature_names is None:
      feature_names = ['v{0}'.format(i) for i in range(len(array[0]))]

    self._array = array
    self._feature_names = feature_names

  def rows(self):
    for ent in self._array:
      yield dict([x for x in zip(self._feature_names, ent) if x[1] is not None])

class ZipArrayLoader(BaseLoader):
  """
  ZipArrayLoader zips multiple 1-d arrays that have the same length.
  Expected to load column-oriented data.

  For example:

  >>> ZipArrayLoader([[1,4], [2,5], [3,6]], ['k1','k2','k3'])

  ... or simply:

  >>> ZipArrayLoader(k1=[1,4], k2=[2,5], k3=[3,6])

  ... will load two entries:

    - {'k1': 1, 'k2': 2, 'k3': 3}
    - {'k1': 4, 'k2': 5, 'k3': 6}
  """

  def __init__(self, arrays=[], feature_names=None, **named_arrays):
    if feature_names is None:
      feature_names = ['v{0}'.format(i) for i in range(len(arrays))]

    if len(arrays) != len(feature_names):
      raise RuntimeError('number of arrays and feature names mismatch')

    self._feature_names = feature_names
    self._arrays = list(arrays)
    for name in named_arrays:
      self._feature_names.append(name)
      self._arrays.append(named_arrays[name])

  def rows(self):
    for ent in zip(*self._arrays):
      yield dict([x for x in zip(self._feature_names, ent) if x[1] is not None])



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import jubatus

from .base import GenericSchema, BaseDataset, BaseService, GenericConfig
from .compat import *

class Schema(GenericSchema):
  """
  Schema for Anomaly service.
  """

  ID = 'i'
  FLAG = 'f'

  def __init__(self, mapping, fallback=None):
    self._id_key = self._get_unique_mapping(mapping, fallback, self.ID, 'ID', True)
    self._flag_key = self._get_unique_mapping(mapping, fallback, self.FLAG, 'FLAG', True)
    super(Schema, self).__init__(mapping, fallback)

  def transform(self, row):
    """
    Anomaly schema transforms the row into Datum, its associated ID and flag.
    Flag can be a value of any type.  It is provided for convenience to
    calculate precision.
    """
    row_id = row.get(self._id_key, None)
    if row_id is not None:
      row_id = unicode_t(row_id)
    row_flag = row.get(self._flag_key, None)
    d = self._transform_as_datum(row, None, [self._id_key, self._flag_key])
    return (row_id, row_flag, d)

class Dataset(BaseDataset):
  """
  Dataset for Anomaly service.
  """

  @classmethod
  def _predict(cls, row):
    return Schema.predict(row, False)

class Anomaly(BaseService):
  """
  Anomaly service.
  """

  @classmethod
  def name(cls):
    return 'anomaly'

  @classmethod
  def _client_class(cls):
    return jubatus.anomaly.client.Anomaly

  def add(self, dataset):
    """
    Adds data points to the anomaly model using the given dataset and returns
    LOF scores.
    """
    cli = self._client()
    for (idx, (row_id, row_flag, d)) in dataset:
      if row_id is not None:
        raise RuntimeError('ID-based datasets must use `overwrite` or `update` instead of `add`')
      result = cli.add(d)
      yield (idx, result.id, row_flag, result.score)

  def update(self, dataset):
    """
    Updates data points in the anomaly model using the given dataset and
    returns LOF scores.
    """
    cli = self._client()
    for (idx, (row_id, row_flag, d)) in dataset:
      if row_id is None:
        raise RuntimeError('Non ID-based datasets must use `add` instead of `update`')
      result = cli.update(row_id, d)
      yield (idx, row_id, row_flag, result)

  def overwrite(self, dataset):
    """
    Overwrites data points in the anomaly model using the given dataset and
    returns LOF scores.
    """
    cli = self._client()
    for (idx, (row_id, row_flag, d)) in dataset:
      if row_id is None:
        raise RuntimeError('Non ID-based datasets must use `add` instead of `overwrite`')
      result = cli.overwrite(row_id, d)
      yield (idx, row_id, row_flag, result)

  def calc_score(self, dataset):
    """
    Calculates LOF scores for the given dataset.
    """
    cli = self._client()
    for (idx, (row_id, row_flag, d)) in dataset:
      result = cli.calc_score(d)
      yield (idx, row_id, row_flag, result)

class Config(GenericConfig):
  """
  Configuration to run Anomaly service.
  """

  @classmethod
  def methods(cls):
    return ['lof', 'light_lof']

  @classmethod
  def _default_method(cls):
    return 'lof'

  @classmethod
  def _default_parameter(cls, method):
    params = {
      'nearest_neighbor_num': 10,
      'reverse_nearest_neighbor_num': 30,
      'method': None,
      'parameter': {},
      'ignore_kth_same_point': True,
    }
    if method == 'lof':
      params['method'] = 'inverted_index_euclid'
    elif method == 'light_lof':
      params['method'] = 'euclid_lsh'
      params['parameter'] = {
        'threads': -1,  # use number of logical CPU cores
        'hash_num': 64,
      }
    else:
      raise RuntimeError('unknown method: {0}'.format(method))
    return params



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

import sys
import logging

# Import log levels.
from logging import DEBUG, INFO, WARNING, ERROR, CRITICAL

# Define the default logging format.
_DEFAULT_FORMAT = '[%(name)s] %(asctime)s: (%(levelname)s) %(message)s'

# jubakit root logger name.
_LOGGER_NAME = 'jubakit'

# Define the defualt logger which does nothing.
class _NullHandler(logging.Handler):
  def emit(self, record):
    pass

_logger = logging.getLogger(_LOGGER_NAME)
_logger.addHandler(_NullHandler())
_logger.setLevel(logging.CRITICAL)

def setup_logger(level=logging.WARNING, f=sys.stderr, log_format=_DEFAULT_FORMAT):
  """
  Convenient method to setup the logger.
  """
  handler = logging.StreamHandler(f)
  handler.setFormatter(logging.Formatter(log_format))

  _logger.propagate = False
  _logger.addHandler(handler)
  _logger.setLevel(level)

def get_logger(name=None):
  """
  Returns the logger.
  If `name` is specified, child logger is returned.
  Otherwise the default jubakit logger is returned.

  This is mainly expected for internal uses but users can get logger
  to print their own logs.
  """
  if name is None:
    return _logger
  elif hasattr(_logger, 'getChild'):
    return _logger.getChild(name)
  else:
    return logging.getLogger('{0}.{1}'.format(_LOGGER_NAME, name))



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

from unittest import TestCase
from tempfile import NamedTemporaryFile as TempFile

from jubakit.loader.csv import CSVLoader

class CSVLoaderTest(TestCase):
  def test_simple(self):
    with TempFile() as f:
      f.write("k1,\"k2\",k3\n1,2,3\n4,5,6".encode('utf-8'))
      f.flush()
      loader = CSVLoader(f.name)
      for row in loader:
        self.assertEqual(set(['k1','k2','k3']), set(row.keys()))
        if row['k1'] == '1':
          self.assertEqual('2', row['k2'])
          self.assertEqual('3', row['k3'])
        elif row['k1'] == '4':
          self.assertEqual('5', row['k2'])
          self.assertEqual('6', row['k3'])
        else:
          self.fail('unexpected row')

  def test_guess_header(self):
    with TempFile() as f:
      f.write("k1,k2,k3\n1,2,3".encode())
      f.flush()
      loader = CSVLoader(f.name, fieldnames=True)
      self.assertEqual([{'k1': '1', 'k2': '2', 'k3': '3'}], list(loader))

  def test_noheader(self):
    with TempFile() as f:
      f.write("1,\"2\",3\n\"4\",5,\"6\"".encode('utf-8'))
      f.flush()
      loader = CSVLoader(f.name, False)
      for row in loader:
        self.assertEqual(set(['c0','c1','c2']), set(row.keys()))
        if row['c0'] == '1':
          self.assertEqual('2', row['c1'])
          self.assertEqual('3', row['c2'])
        elif row['c0'] == '4':
          self.assertEqual('5', row['c1'])
          self.assertEqual('6', row['c2'])
        else:
          self.fail('unexpected row')

  def test_cp932(self):
    with TempFile() as f:
      f.write("テスト1,テスト2".encode('cp932'))
      f.flush()
      loader = CSVLoader(f.name, None, 'cp932')
      for row in loader:
        self.assertEqual('テスト1', row['c0'])
        self.assertEqual('テスト2', row['c1'])



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

from ..base import BaseLoader
from ..compat import *

"""
Chain loaders loads records from other loader(s).
"""

class MergeChainLoader(BaseLoader):
  """
  MergeChainLoader merges multiple loaders.
  """

  def __init__(self, *loaders):
    self._loaders = loaders

  def rows(self):
    for ent in zip_longest(*self._loaders, fillvalue={}):
      merged = {}
      for d in ent:
        merged.update(d)
      yield merged

class ValueMapChainLoader(BaseLoader):
  """
  ValueMapChainLoader is a loader to map value of the specified key in each
  record loaded from another loader.
  """

  def __init__(self, loader, key, mapping):
    self._loader = loader
    self._key = key
    self._mapping = mapping

  def rows(self):
    for ent in self._loader:
      ent[self._key] = self._mapping[ent[self._key]]
      yield ent



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

from unittest import TestCase

from jubakit.loader.twitter import TwitterStreamLoader, TwitterOAuthHandler

class TwitterStreamLoaderTest(TestCase):
  oauth = TwitterOAuthHandler(consumer_key='x', consumer_secret='x', access_token='x', access_secret='x')

  def test_simple(self):
    loader = TwitterStreamLoader(self.oauth)

  def test_errors(self):
    # auth info (both os.environ and auth argument) is not set
    self.assertRaises(RuntimeError, TwitterStreamLoader)

    # invalid stream name
    self.assertRaises(RuntimeError, TwitterStreamLoader, self.oauth, 'invalid_mode')



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

from ..base import BaseLoader
from ..compat import *

class SparseMatrixLoader(BaseLoader):
  """
  SparseMatrixLoader is a loader to read from scipy.sparse 2-d matrix.
  Zero entries are ignored.
  """

  def __init__(self, matrix, feature_names=None):
    self._matrix = matrix.tocsr()
    self._feature_names = feature_names

  def rows(self):
    m = self._matrix
    for i in range(m.shape[0]):
      cols = m.indices[m.indptr[i]:m.indptr[i+1]]

      if self._feature_names is None:
        fv_names = ['v{0}'.format(col) for col in cols]
      else:
        fv_names = [self._feature_names[col] for col in cols]

      data = dict(zip(fv_names, m.data[m.indptr[i]:m.indptr[i+1]]))
      yield data



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup, find_packages

def _read(filename):
  with open(filename) as f:
    return f.read()

# Load package version.
exec(_read('jubakit/_version.py'))

setup(name='jubakit',
      version='.'.join(str(x) for x in VERSION),
      description='Jubatus Toolkit',
      long_description=_read('README.rst'),
      url='http://jubat.us',
      author='PFN & NTT',
      author_email='jubatus-team@googlegroups.com',
      license='MIT',
      classifiers=[
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Programming Language :: Python :: 3.5',
      ],
      packages=find_packages(exclude=['jubakit.test']),
      install_requires=[
          'jubatus>=0.8.0',
          'psutil',
      ],
      extras_require={
          'test': ['numpy', 'scipy', 'scikit-learn', 'tweepy', 'jq'],
      },
      test_suite = 'jubakit.test',
)



# -*- coding: utf-8 -*-

from __future__ import absolute_import, division, print_function, unicode_literals

from unittest import TestCase

from jubakit.loader.array import ArrayLoader
from jubakit.loader.chain import MergeChainLoader, ValueMapChainLoader

class MergeChainLoaderTest(TestCase):
  def test_simple(self):
    loader = MergeChainLoader(
      ArrayLoader([[0,1],[2,3],[4,5]], ['v1','v2']),
      ArrayLoader([[0,1],[2,3],[4,5]], ['v3','v4']),
    )
    for row in loader:
      self.assertEqual(set(['v1','v2','v3','v4']), set(row.keys()))
      if row['v1'] == 0:
        self.assertEqual(1, row['v2'])
        self.assertEqual(0, row['v3'])
        self.assertEqual(1, row['v4'])
      elif row['v1'] ==2:
        self.assertEqual(3, row['v2'])
        self.assertEqual(2, row['v3'])
        self.assertEqual(3, row['v4'])
      elif row['v1'] == 4:
        self.assertEqual(5, row['v2'])
        self.assertEqual(4, row['v3'])
        self.assertEqual(5, row['v4'])
      else:
        self.fail('unexpected row: {0}'.format(row))

class ValueMapChainLoaderTest(TestCase):
  def test_simple(self):
    loader = ValueMapChainLoader(
      ArrayLoader([[0,1],[2,3],[4,5]], ['v1','v2']),
      'v2',
      {1: '_test1', 3: '_test3', 5: '_test5'}
    )
    for row in loader:
      self.assertEqual(set(['v1','v2']), set(row.keys()))
      if row['v1'] == 0:
        self.assertEqual('_test1', row['v2'])
      elif row['v1'] == 2:
        self.assertEqual('_test3', row['v2'])
      elif row['v1'] == 4:
        self.assertEqual('_test5', row['v2'])
      else:
        self.fail('unexpected row: {0}'.format(row))
